import requests
from tqdm import tqdm

url = 'https://pku-milab-model.oss-ap-northeast-1.aliyuncs.com/fold0.pth'
response = requests.get(url, stream=True)
total_size_in_bytes = int(response.headers.get('content-length', 0))
with open('weights/fold0.pth', 'wb') as file:
    with tqdm(total=total_size_in_bytes, unit='B', unit_scale=True) as bar:
        # 分块下载文件
        for chunk in response.iter_content(chunk_size=1024):
            file.write(chunk)
            bar.update(len(chunk))  # 更新进度条
url = 'https://pku-milab-model.oss-ap-northeast-1.aliyuncs.com/fold1.pth'
response = requests.get(url, stream=True)
total_size_in_bytes = int(response.headers.get('content-length', 0))
with open('weights/fold1.pth', 'wb') as file:
    with tqdm(total=total_size_in_bytes, unit='B', unit_scale=True) as bar:
        # 分块下载文件
        for chunk in response.iter_content(chunk_size=1024):
            file.write(chunk)
            bar.update(len(chunk))  # 更新进度条
url = 'https://pku-milab-model.oss-ap-northeast-1.aliyuncs.com/fold2.pth'
response = requests.get(url, stream=True)
total_size_in_bytes = int(response.headers.get('content-length', 0))
with open('weights/fold2.pth', 'wb') as file:
    with tqdm(total=total_size_in_bytes, unit='B', unit_scale=True) as bar:
        # 分块下载文件
        for chunk in response.iter_content(chunk_size=1024):
            file.write(chunk)
            bar.update(len(chunk))  # 更新进度条



from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from xgboost import XGBClassifier
from sklearn.metrics import (accuracy_score, classification_report, 
                             roc_auc_score, roc_curve, auc)
from sklearn.preprocessing import label_binarize
from sklearn.multiclass import OneVsRestClassifier
import matplotlib.pyplot as plt
import numpy as np

class Classifier:
    def __init__(self, model_name='xgboost',classnum=2, **model_params):
        self.ERROR_ICON = '❌'
        self.SUCCESS_ICON = '✅'
        self.RUNNING_ICON = '⏳'
        self.INFO_ICON = '➤'
        self.model_name = model_name.lower()
        self.model = None
        self.classes_ = None
        self._init_model(classnum=classnum, **model_params)
        self.classnum = classnum

    def _init_model(self, classnum, **params):
        """初始化支持多分类的模型"""
        model_config = {
            'xgboost': {
                'class': XGBClassifier,
                'default': {
                    'n_estimators': 100,
                    'random_state': 42,
                    'objective': 'binary:logistic' if classnum == 2 else 'multi:softmax',
                }
            },
            'rf': {
                'class': RandomForestClassifier,
                'default': {
                    'n_estimators': 100,
                    'random_state': 42,
                    'class_weight': 'balanced'
                }
            },
            'svm': {
                'class': SVC,
                'default': {
                    'probability': True,
                    'random_state': 42,
                    'decision_function_shape': 'ovr' if classnum > 2 else 'ovo'
                }
            },
            'lr': {
                'class': LogisticRegression,
                'default': {
                    'max_iter': 1000,
                    'random_state': 42,
                    'multi_class': 'multinomial' if classnum > 2 else 'auto',
                    'solver': 'lbfgs' if classnum > 2 else 'liblinear'
                }
            },
            'knn': {
                'class': KNeighborsClassifier,
                'default': {
                    'n_neighbors': 5,
                    'weights': 'distance'
                }
            }
        }
        
        if self.model_name not in model_config:
            raise ValueError(f"{self.ERROR_ICON} 不支持模型类型，可选: {list(model_config.keys())}")
            
        config = model_config[self.model_name]
        final_params = {
            **config['default'],
            **{k: v for k, v in params.items() if k not in ['classnum']}
        }
        
        if classnum > 2 and self.model_name in ['svm', 'lr']:
            if 'OneVsRest' not in str(config['class']):
                self.model = OneVsRestClassifier(config['class'](**final_params))
        else:
            self.model = config['class'](**final_params)
            
        print(f"{self.SUCCESS_ICON} {self.model_name.upper()} 分类器初始化成功")

    def fit(self, X_train, y_train):
        """训练流程增强"""
        self.classes_ = np.unique(y_train)
        
        if len(self.classes_) != self.classnum:
            raise ValueError(
                f"{self.ERROR_ICON} 类别数不匹配！输入数据有{len(self.classes_)}类，但设置classnum={self.classnum}"
            )
        
        if self.model_name == 'xgboost':
            if self.classnum > 2 and 'num_class' not in self.model.get_params():
                self.model.set_params(num_class=self.classnum)
            elif self.classnum == 2 and self.model.objective.startswith('multi'):
                self.model.set_params(objective='binary:logistic')
            
        print(f"{self.RUNNING_ICON} 开始训练... (类别数: {len(self.classes_)})")
        self.model.fit(X_train, y_train)
        print(f"{self.SUCCESS_ICON} 训练完成")
        return self
    
    def predict(self, X_test):
        print(f"{self.RUNNING_ICON} 生成预测...")
        X_test = np.array(X_test)
        return self.model.predict(X_test)
    
    def predict_proba(self, X_test):
        """获取类别概率"""
        if hasattr(self.model, 'predict_proba'):
            return self.model.predict_proba(X_test)
        raise ValueError(f"{self.ERROR_ICON} 该模型不支持概率预测")

    def evaluate(self, X_test, y_test):
        """多分类评估指标"""
        y_pred = self.predict(X_test)
        y_prob = self.predict_proba(X_test)
        
        # 基础指标
        acc = accuracy_score(y_test, y_pred)
        report = classification_report(y_test, y_pred)
        
        # ROC AUC计算
        if len(self.classes_) == 2:
            roc_auc = roc_auc_score(y_test, y_prob[:, 1])
        else:
            y_test_bin = label_binarize(y_test, classes=self.classes_)
            roc_auc = roc_auc_score(y_test_bin, y_prob, multi_class='ovr')
        
        print(f"\n{self.INFO_ICON} 分类报告:")
        print(report)
        print(f"{self.INFO_ICON} ROC AUC: {roc_auc:.4f}")
        return {'accuracy': acc, 'roc_auc': roc_auc}

    def plot_roc_curve(self, X_test, y_test,category_name,savepath=None):
        """绘制多类别ROC曲线"""
        y_prob = self.predict_proba(X_test)
        n_classes = len(self.classes_)
        
        # 二分类处理
        if n_classes == 2:
            fpr, tpr, _ = roc_curve(y_test, y_prob[:, 1])
            roc_auc = auc(fpr, tpr)
            
            plt.plot(fpr, tpr, label=f'ROC curve (AUC = {roc_auc:.2f})')
        # 多分类处理
        else:
            y_test_bin = label_binarize(y_test, classes=self.classes_)
            fpr = dict()
            tpr = dict()
            roc_auc = dict()
            
            for i in range(n_classes):
                fpr[i], tpr[i], _ = roc_curve(y_test_bin[:, i], y_prob[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])
            
            # 绘制每个类别的曲线
            for i in range(n_classes):
                plt.plot(fpr[i], tpr[i], 
                         label=f'{category_name[i]} (AUC = {roc_auc[i]:.2f})')
            
            # 计算宏平均
            all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
            mean_tpr = np.zeros_like(all_fpr)
            for i in range(n_classes):
                mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])
            mean_tpr /= n_classes
            macro_auc = auc(all_fpr, mean_tpr)
            plt.plot(all_fpr, mean_tpr, 'k--', 
                     label=f'Macro-average (AUC = {macro_auc:.2f})')

        plt.plot([0, 1], [0, 1], 'r--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Multi-class ROC Curve')
        plt.legend(loc="lower right")
        if savepath is not None:
            plt.savefig(savepath,dpi=300,bbox_inches='tight')
        plt.show()


from metabolism.Helper import Helper
from metabolism.CorrelationAnalyzer import CorrelationAnalyzer
import SimpleITK as sitk
import numpy as np
import os,json
from pathlib import Path
import pandas as pd
from scipy import stats, linalg

def resample_image(input_image, reference_image,default=0):
    resampler = sitk.ResampleImageFilter()
    resampler.SetReferenceImage(reference_image)
    
    resampler.SetInterpolator(sitk.sitkNearestNeighbor)
    
    resampler.SetOutputSpacing(reference_image.GetSpacing())
    resampler.SetOutputOrigin(reference_image.GetOrigin())
    resampler.SetOutputDirection(reference_image.GetDirection())
    
    resampler.SetSize(reference_image.GetSize())
    resampler.SetDefaultPixelValue(default)
    return resampler.Execute(input_image)


class StabilityTester:
    def __init__(self):
        self.ERROR_ICON = '❌'
        self.SUCCESS_ICON = '✅'
        self.RUNNING_ICON = '⏳'
        self.INFO_ICON = '➤'
        print(f"[{self.INFO_ICON}] This class aims to find the number of reference group and control group for perturbed methods !!")

        self.correlationAnalyzer = CorrelationAnalyzer()
        self.helper = Helper()
    def suvseg2feature(self,dataset,modelname):
        print(f"[{self.INFO_ICON}] Load SUV and SEG from :", dataset)
        infos = pd.read_csv(os.path.join(dataset,"info.csv"))
        for file in os.listdir(dataset):
            if "SUV" not in file or ".nii.gz" not in file:
                continue
            print("-"*20)
            suv = os.path.join(dataset,file)
            segpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),"merge.nii.gz")
            if not os.path.exists(segpath):
                print(f"[{self.ERROR_ICON}] No segment result for Model {modelname}\n    SUV path:{suv}")
                continue
            print(f" [{self.RUNNING_ICON }] Processing: {suv}")
            suv = sitk.ReadImage(suv)
            seg = sitk.ReadImage(segpath)
            suv = resample_image(suv,seg,default=0)
            featpath = f"{Path(segpath).parent}/{modelname}#suvr.json"
            feat = self.extract_suv_based_roi(suv,seg,modelname=modelname)
            print("#".join(file.replace(".nii.gz","").split("#")[1:]))
            tmp = infos[infos.name == "#".join(file.replace(".nii.gz","").split("#")[1:])]
            gender = tmp.Gender.values[0]
            try: 
                age = float(tmp.Age.values[0])
            except:
                age = "Unknown"
            try:
                weight = float(tmp.Weight.values[0])
            except:
                age = "Unknown"
            feat.update({'gender':gender,"age":age,"weight":weight})
            with open(featpath,"w") as f:
                json.dump(feat,f)
            print(f" [{self.SUCCESS_ICON}SUCCESS] roi features are written into: {featpath}")
    def analyze(self,dataset,modelname):
        print(f"[{self.INFO_ICON}] Load FEAT from :", dataset)
        df = self.correlationAnalyzer.dataset2df(dataset,modelname)
        result = self.helper.detailed_nan_inspection(df)
        if result["msg"] == "error":
            print(f"[{self.ERROR_ICON} WRONG] Feature has NAN or INF")
        else:
            print(f"[{self.SUCCESS_ICON} SUCCESS] Feature has no NAN or INF")
        print('-'*20)
        print(f"[{self.INFO_ICON}] Starting stability Test. Data Count:{len(df)}")
        self.full_refnetwork = self.correlationAnalyzer.compute_partial_correlation_matrix(df)
        results = {}
        for n in range(10,len(df)+1):
            mean_,std_ = self.bootstrap_pcorr_stability(df,resample_size=n,n_iter=100)
            results[n] = (mean_,std_)
            print(f" [{self.SUCCESS_ICON}] resample size:{n}, correlation coefficient:{mean_}")
        return df,results
    
    def bootstrap_pcorr_stability(self,df, resample_size, n_iter=20, seed=42):
        np.random.seed(seed)
        covar_cols = ['Age', 'Sex', 'Weight']  # 根据实际情况修改
        all_pcorr_flatten = []
        correlations = []
        for _ in range(n_iter):
            idx = np.random.choice(df.index,size=resample_size,replace=False)
            sub_df = df.loc[idx].copy()
            sub_refnetwork = self.correlationAnalyzer.compute_partial_correlation_matrix(sub_df,z=['Age', 'Sex', 'Weight'])
            
            flat_full = self.full_refnetwork[~np.eye(self.full_refnetwork.shape[0], dtype=bool)]
            flat_sub = sub_refnetwork[~np.eye(sub_refnetwork.shape[0], dtype=bool)]
            r = np.corrcoef(flat_full, flat_sub)[0,1]
            correlations.append(r)
        return np.mean(correlations), np.std(correlations)
    
    def extract_suv_based_roi(self,suv,seg,modelname):
        seg = sitk.GetArrayFromImage(seg)
        suv = sitk.GetArrayFromImage(suv)
        if modelname == "mpum":
            atlasinfo = self.helper.get_mpum_categories()
            brainzone= [x for x in range(132,215)]
        op = {}
        mask = np.isin(seg, brainzone)
        meanAllBrainSUV = np.mean(suv[mask])
        for i in range(0,len(atlasinfo)):
            tmp = suv[seg == i]
            if len(tmp) == 0:
                continue
            meansuvr = np.mean(tmp) / meanAllBrainSUV
            maxsuvr = np.max(tmp) / meanAllBrainSUV
            op[atlasinfo[str(i)]] = {"meansuvr":float(meansuvr),"maxsuvr":float(maxsuvr)}
        return op


from metabolism.CorrelationAnalyzer import CorrelationAnalyzer
from metabolism.Helper import Helper
import numpy as np
import pandas as pd
from scipy.stats import norm
import SimpleITK as sitk
from scipy.stats import gaussian_kde
import os,time
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor

class Network:
    def __init__(self):
        self.ERROR_ICON = '❌'
        self.SUCCESS_ICON = '✅'
        self.RUNNING_ICON = '⏳'
        self.INFO_ICON = '➤'
        self.correlationAnalyzer = CorrelationAnalyzer()
        self.helper = Helper()
    def process(self,dataset1,dataset2,modelname):
        pass
class GroupLevelNetwork(Network):
    def process_pcorr_diffgroup(self,dataset1,dataset2,modelname):
        pcorr1,df1 = self.dataset2pcorr(dataset1,modelname)
        pcorr2,df2 = self.dataset2pcorr(dataset2,modelname)
        numerator = pcorr1 - pcorr2
        denominator = pcorr1 + pcorr2
        denominator[denominator == 0] = 1e-6
        diffgroup = numerator / denominator
        print(f"[{self.SUCCESS_ICON}] group-level metabolic network DIFF is returned !!")
        return diffgroup
    def dataset2pcorr(self,dataset,modelname):
        df = self.correlationAnalyzer.dataset2df(dataset,modelname)
        print(f"[{self.INFO_ICON}] dataset:",dataset)
        print(f"[{self.SUCCESS_ICON}] dataset count:{len(df)}")
        pcorr = self.correlationAnalyzer.compute_partial_correlation_matrix(df,z=['Age', 'Sex', 'Weight'])
        return pcorr,df
    def dataset2pearson(self,dataset,modelname):
        df = self.correlationAnalyzer.dataset2df(dataset,modelname)
        df_ = df.drop(columns = ["Age","Sex","Weight"])
        print(f"[{self.INFO_ICON}] dataset:",dataset)
        print(f"[{self.SUCCESS_ICON}] dataset count:{len(df)}")
        corr,roi_name = self.correlationAnalyzer.compute_correlation_matrix(df_)
        return corr,roi_name

class IndividualNetwork(Network):
    def __init__(self,control,dataset2,modelname):
        super().__init__()
        print("="*20,"This method need one suv to create network !!")
        self.dataset1 = control
        self.dataset2 = dataset2
        self.modelname = modelname
    def process(self,referencenumber,covariate=['Age', 'Sex', 'Weight']):
        dataset1 = self.dataset1
        dataset2 = self.dataset2
        modelname = self.modelname
        df1 = self.correlationAnalyzer.dataset2df(dataset1,modelname)
        df2 = self.correlationAnalyzer.dataset2df(dataset2,modelname)
        self.df1 = df1
        self.df2 = df2
        print(f"[{self.INFO_ICON}] dataset1:",dataset1)
        print(f"[{self.SUCCESS_ICON}] dataset1 count:{len(df1)}")
        print(f"[{self.INFO_ICON}] dataset2:",dataset2)
        print(f"[{self.SUCCESS_ICON}] dataset2 count:{len(df2)}")
        df1_referece,df1_control = self.helper.split_dataframe(df1,referencenumber,random_seed=42)
        refnet = self.correlationAnalyzer.compute_partial_correlation_matrix(df1_referece,z=['Age', 'Sex', 'Weight'])

        edge_significant_counts = np.zeros_like(refnet, dtype=int)
        df2_zscores = []
        for i, row_dict in df2.iterrows():
            df1_referece_ = df1_referece.copy()
            current_row = pd.DataFrame([row_dict])
            combined = pd.concat([df1_referece_, current_row], ignore_index=True)
            # Perturbed network
            pnet = self.correlationAnalyzer.compute_partial_correlation_matrix(combined,z=['Age', 'Sex', 'Weight'])
            delta_pnet = pnet - refnet
            se = (1 - refnet**2) / (referencenumber - len(covariate) - 1)
            z_scores = delta_pnet / se
            df2_zscores.append(z_scores)

            p_values = 2 * norm.sf(np.abs(z_scores))
            num_edges = p_values.size - np.isnan(p_values).sum()
            alpha_corrected = 0.05 / num_edges
            significant_edges = p_values < alpha_corrected
            edge_significant_counts[significant_edges] += 1

        control_zscores = []
        for i, row_dict in df1_control.iterrows():
            df1_referece_ = df1_referece.copy()
            current_row = pd.DataFrame([row_dict])
            combined = pd.concat([df1_referece_, current_row], ignore_index=True)
            # Perturbed network
            pnet = self.correlationAnalyzer.compute_partial_correlation_matrix(combined,z=['Age', 'Sex', 'Weight'])
            delta_pnet = pnet - refnet
            se = (1 - refnet**2) / (referencenumber - len(covariate) - 1)
            z_scores = delta_pnet / se
            control_zscores.append(z_scores)
        self.control_zscores = control_zscores
        self.df2_zscores = df2_zscores
        self.edge_significant_counts = edge_significant_counts
        print(f"[{self.SUCCESS_ICON}SUCCESS] return dataset: control_zscores,df2_zscores,edge_significant_counts")
        return control_zscores,df2_zscores,edge_significant_counts
    def analyze_multiorgan(self):
        print(f"[{self.INFO_ICON}] multiorgan analyze starting")
        edge_significant_counts = self.edge_significant_counts
        df2_zscores = self.df2_zscores
        control_zscores = self.control_zscores
        
        mask_lower = np.triu(np.ones_like(edge_significant_counts, dtype=bool), k=1)
        edge_significant_counts[mask_lower] = 0
        rows, cols = np.where(edge_significant_counts != 0)
        values = self.edge_significant_counts[rows, cols]
        edges = list(zip(values, rows, cols))
        sorted_edges = sorted(edges, key=lambda x: x[0], reverse=True)

        columns = [x for x in self.df1.columns if x not in ['Age', 'Sex', 'Weight']]
        control_zscore_mean = np.mean(control_zscores,axis=0)
        df2_zscore_mean = np.mean(df2_zscores,axis=0)
        control_zscore_std = np.std(control_zscores,axis=0)
        df2_zscore_std = np.std(df2_zscores,axis=0)
        data_list = []
        
        for count,row,col in sorted_edges:
            if count < len(self.df2) * 0.2:
                continue
            roi1 = columns[row]
            roi2 = columns[col]
            meanz1 = df2_zscore_mean[row,col]
            stdz1 = df2_zscore_std[row,col]
            meanz2 = control_zscore_mean[row,col]
            stdz2 = control_zscore_std[row,col]
            data_list.append({"SignificantCount":count,"roi1":roi1,"roi2":roi2,"patient_z_mean":meanz1,
                 "patient_z_std":stdz1,"control_z_mean":meanz2,"control_z_std":stdz2})
        df = pd.DataFrame(data_list)
        print(f"[{self.SUCCESS_ICON}SUCCESS] return dataframe: SignificantCount,roi1,roi2,patient_z_mean,patient_z_std,control_z_mean,control_z_std")
        return df

    def analyze_singleorgan(self):
        print(f"[{self.INFO_ICON}] singleorgan analyze starting...")
        df2_zscores = self.df2_zscores
        control_zscores = self.control_zscores
        control_zscores = np.abs(np.array(control_zscores))
        df2_zscores = np.abs(np.array(df2_zscores))
        
        columns = [x for x in self.df1.columns if x not in ['Age', 'Sex', 'Weight']]
        data_list = []
        for i in range(len(columns)):
            roi = columns[i]
            control_degree = np.sum(control_zscores[:,i],axis=1)
            df2_degree = np.sum(df2_zscores[:,i],axis=1)
            df2_degree_max = np.max(df2_degree)
            df2_degree_mean = np.mean(df2_degree)
            df2_degree_std = np.std(df2_degree)
            control_degree_max = np.max(control_degree)
            control_degree_mean = np.mean(control_degree)
            control_degree_std = np.std(control_degree)
            data_list.append({"roi":roi,"patient_degree_max":df2_degree_max,"patient_degree_mean":df2_degree_mean,"patient_degree_std":df2_degree_std,
                              "control_degree_max":control_degree_max,"control_degree_mean":control_degree_mean,"control_degree_std":control_degree_std})
        df = pd.DataFrame(data_list)
        df = df.sort_values(by="patient_degree_mean", ascending=False)
        print(f"[{self.SUCCESS_ICON}SUCCESS] return dataframe: roi,patient_degree_max,patient_degree_mean,control_degree_max,control_degree_mean")
        return df

class KLNetwork(Network):
    def __init__(self,dataset,modelname):
        super().__init__()
        print("="*20,"\nThis method need one suv to create network !!\n","="*20)
        self.dataset = dataset
        self.modelname = modelname
    def process(self):
        dataset = self.dataset
        modelname = self.modelname
        print(f"[{self.INFO_ICON}] dataset:",dataset)
        for file in os.listdir(dataset):
            if "SUV" not in file or ".nii.gz" not in file:
                continue
            suv = os.path.join(dataset,file)
            segpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),"merge.nii.gz")
            npzpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),"KLSE_matrix.npz")
            print(suv)
            print(segpath)
            if not os.path.exists(segpath):
                print(f"[{self.ERROR_ICON}] No segment result for Model {modelname}\n    SUV path:{suv}")
                continue
            if os.path.exists(npzpath):
                print(f" [{self.SUCCESS_ICON} SKIP] Result has saved to:",npzpath)
                continue
            suv = sitk.ReadImage(suv)
            seg = sitk.ReadImage(segpath)
            # step1: get flatten value from each roi
            roi_intensities,roi_names = self.extract_roi_intensities(suv,seg)
            # step2: compute klse martix
            print(f" [{self.RUNNING_ICON}] Using KDE to estimate KLNetwork ...")
            # klse = self.compute_klse_matrix(roi_intensities)
            klse = self.compute_klse_matrix_parallel(roi_intensities)
            
            
            print(f" [{self.SUCCESS_ICON}] Estimation Success !! categoriy:{len(roi_names)}")
            roi_names = np.array(roi_names, dtype=object)
            np.savez_compressed(npzpath, 
                               klse_matrix=klse, 
                               roi_names=roi_names)
            print(f" [{self.SUCCESS_ICON}] Result saved to:",npzpath)

    def extract_roi_intensities(self,suv, seg):
        modelname = self.modelname
        seg = sitk.GetArrayFromImage(seg)
        suv = sitk.GetArrayFromImage(suv)
        if modelname == "mpum":
            atlasinfo = self.helper.get_mpum_categories()
            brainzone = [x for x in range(132,215)]
        roi_intensities = []
        roi_names = []
        for label in brainzone:
            mask = (seg == label)
            intensities = suv[mask].flatten()
            if label == 0:
                continue
            if len(intensities) == 0:
                continue
            roi_names.append(atlasinfo[str(label)])
            roi_intensities.append(intensities)
        return roi_intensities,roi_names
    def compute_klse_matrix(self,roi_intensities):
        """计算对称KL散度并生成KLSE矩阵"""
        epsilon=1e-8
        n_roi = len(roi_intensities)
        klse_matrix = np.zeros((n_roi, n_roi))
        
        # 为每个ROI构建KDE模型
        kdes = [gaussian_kde(roi) for roi in roi_intensities]
        
        for i in tqdm(range(n_roi)):
            for j in range(i, n_roi):
                if i == j:
                    klse_matrix[i,j] = 1.0  # 对角线元素为自连接
                    continue
                    
                # 获取KDE概率密度函数
                kde_p = kdes[i]
                kde_q = kdes[j]
                
                # 定义积分域（覆盖两个分布的联合范围）
                x_min = min(kde_p.dataset.min(), kde_q.dataset.min())
                x_max = max(kde_p.dataset.max(), kde_q.dataset.max())
                x = np.linspace(x_min, x_max, 1000)
                
                # 计算概率密度（防止零值）
                p = np.clip(kde_p(x), epsilon, None)
                q = np.clip(kde_q(x), epsilon, None)
                
                # 计算对称KL散度
                kl_pq = np.trapz(p * np.log(p/q), x)
                kl_qp = np.trapz(q * np.log(q/p), x)
                sym_kl = kl_pq + kl_qp
                
                # 计算KLSE连接强度
                klse = np.exp(-sym_kl)
                klse_matrix[i,j] = klse
                klse_matrix[j,i] = klse  # 对称赋值
        return klse_matrix

    def compute_klse_matrix_parallel(self, roi_intensities):
        n_roi = len(roi_intensities)
        klse_matrix = np.zeros((n_roi, n_roi))
        kdes = []
        for roi in roi_intensities:
            if len(roi) < 2:
                roi = np.repeat(roi, 2) + np.random.normal(0, 1e-6, size=2)
            kdes.append(gaussian_kde(roi))
        self.kdes = kdes
        # 预计算所有KDE的范围 [优化点1]
        self.mins = [kde.dataset.min() for kde in kdes]
        self.maxs = [kde.dataset.max() for kde in kdes]
        
        # 生成所有需要计算的(i,j)对 [优化点2]
        pairs = [(i, j) for i in range(n_roi) for j in range(i, n_roi)]
        # 启动多进程池
        with ProcessPoolExecutor(max_workers=int(16)) as executor:
            results = list(tqdm(executor.map(self._compute_pair, pairs), total=len(pairs)))
        
        # 填充矩阵
        for (i, j), val in results:
            klse_matrix[i, j] = klse_matrix[j, i] = val
        
        return klse_matrix
    def _compute_pair(self,pair):
        i, j = pair
        mins = self.mins
        maxs = self.maxs
        kdes = self.kdes
        if i == j:
            return (i, j), 1.0
        # 计算积分域
        x_min = min(mins[i], mins[j])
        x_max = max(maxs[i], maxs[j])
        x = np.linspace(x_min, x_max, 100)  # 减少采样点 [优化点3]
        # 向量化评估KDE [优化点4]
        p = np.clip(kdes[i](x), 1e-8, None)
        q = np.clip(kdes[j](x), 1e-8, None)
        # 快速积分计算
        sym_kl = np.trapz(p * np.log(p/q), x) + np.trapz(q * np.log(q/p), x)
        return (i, j), np.exp(-sym_kl)
    
    def analyze(self):
        dataset = self.dataset
        modelname = self.modelname
        print(f"[{self.INFO_ICON}] dataset:",dataset)

        all_roi_names = []
        valid_files = []

        for file in os.listdir(dataset):
            if "SUV" not in file or ".nii.gz" not in file:
                continue
            npzpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),"KLSE_matrix.npz")
            if not os.path.exists(npzpath):
                continue
            data = np.load(npzpath, allow_pickle=True)
            roi_names = data["roi_names"].tolist()
            all_roi_names.append(roi_names)
            valid_files.append(npzpath)

        if not valid_files:
            print(f"[{self.ERROR_ICON}] No valid KLSE files. Run .process() or check dataset path !!")
            return None,None
        else:
            print(f"[{self.INFO_ICON}] Find {len(valid_files)} KLSE data !!")

        common_roi = []
        if all_roi_names:
            base_names = all_roi_names[0]
            common_roi = [
                name for name in base_names 
                if all(name in names for names in all_roi_names[1:])
            ]
        print(f"[{self.SUCCESS_ICON}] Effective Region Count: {len(common_roi)}")
        
        processed_klses = []
        removed_records = []
        
        for file_idx, npz_path in enumerate(valid_files):
            data = np.load(npz_path, allow_pickle=True)
            original_klse = data["klse_matrix"]
            original_names = data["roi_names"].tolist()
            
            name_to_idx = {name: idx for idx, name in enumerate(original_names)}
            
            try:
                keep_indices = [name_to_idx[name] for name in common_roi]
            except KeyError as e:
                print(f" [{self.ERROR_ICON}] 文件 {npz_path} 缺失关键脑区: {e}")
                continue
            
            row_indices = np.array(keep_indices)[:, np.newaxis]
            col_indices = np.array(keep_indices)
            
            processed_klse = original_klse[row_indices, col_indices]
            
            removed = [name for name in original_names if name not in common_roi]
            if removed:
                print(f" [{self.SUCCESS_ICON}]  Remove region: {', '.join(removed)} from {npz_path}")
                removed_records.append((os.path.basename(npz_path), removed))
            
            processed_klses.append(processed_klse)
        
        if len(common_roi) == 0:
            print(f" [{self.ERROR_ICON}] No Common region ?!")
            return None, None
        
        self.klses = processed_klses
        self.common_roi_names = common_roi
        return processed_klses, common_roi


import SimpleITK as sitk
import numpy as np
import os,json
from pathlib import Path
import pandas as pd
from scipy import stats, linalg

class CorrelationAnalyzer:
    def __init__(self):
        self.ERROR_ICON = '❌'
        self.SUCCESS_ICON = '✅'
        self.RUNNING_ICON = '⏳'
        self.INFO_ICON = '➤'
        pass
    def dataset2df(self,dataset,modelname):
        feats = []
        for file in os.listdir(dataset):
            if "SUV" not in file or ".nii.gz" not in file:
                continue
            suv = os.path.join(dataset,file)
            segpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),"merge.nii.gz")
            if not os.path.exists(segpath):
                print(f"[{self.ERROR_ICON}] No segment result for Model {modelname}\n    SUV path:{suv}")
                continue
            featpath = os.path.join(dataset,"{}#{}".format(modelname,file.replace(".nii.gz","")),f"{modelname}#suvr.json")
            with open(featpath,"r") as f:
                feat = json.load(f)
            feats.append(feat)
        df = self._preprocess_feature(feats)
        return df
    def _preprocess_feature(self,feats):
        processed = []
        for feat in feats:
            try:
                subject_data = {
                    'Age': int(feat['age']),  # 转换字符串年龄为整数
                    'Sex': 0 if feat['gender'] == 'M' else 1,  # 性别编码
                    'Weight': feat['weight']
                }
                for key in feat:
                    if key in ['background', 'gender', 'age', 'weight','L-Substantia nigra','R-Substantia nigra']:
                        continue  #
                    subject_data[key] = feat[key]['meansuvr']
                processed.append(subject_data)
            except:
                pass
        return pd.DataFrame(processed)
    def compute_correlation_matrix(self,df):
        corr_matrix = df.corr(method='pearson')
        return corr_matrix.values,corr_matrix.columns.values
    def compute_partial_correlation_matrix(self,df,z=['Age', 'Sex', 'Weight']):
        '''partial correlation matrix: parital correlation between X and Y after removing linear effects of Z'''
        regions = [col for col in df.columns if col not in z]
        X = df[regions].values
        cover = []
        for zz in z:
            cover.append(df[zz].values)
        cover.append(np.ones(len(df)))
        covar = np.column_stack(cover)
        
        covar = (covar - covar.mean(axis=0)) / (covar.std(axis=0) + 1e-8)
        residuals = []
        for i in range(X.shape[1]):
            try:
                beta = linalg.lstsq(covar, X[:,i], cond=1e-6)[0]
            except:
                beta = np.dot(np.linalg.pinv(covar), X[:,i])
            resid = X[:,i] - np.dot(covar, beta)
            residuals.append(resid)
        
        residuals = np.array(residuals).T
        residuals += np.random.normal(0, 1e-8, residuals.shape)
        pcorr = np.corrcoef(residuals, rowvar=False)
        np.fill_diagonal(pcorr, 0)
        return pcorr